from PIL import Image

""" test fixtures """

import os

import numpy as np
import pytest

from ffpiv import sample_data, window


@pytest.fixture()
def path_img():
    """Path to sample image files."""
    return os.path.join(
        os.path.dirname(__file__),
        "..",
        "examples",
    )


@pytest.fixture()
def fns_img():
    """Collect image files."""
    return sample_data.get_hommerich_files()


@pytest.fixture()
def imgs(fns_img):
    """4 selected frames from sample dataset, read with reader helper function.

    Result is [4 x n x m ] np.ndarray
    """
    return np.stack([np.array(Image.open(fn)) for fn in fns_img[0:4]])


@pytest.fixture()
def imgs_win(imgs):
    """Prepare stack of interrogation windows per image in stack of images."""
    # get the x and y coordinates per window
    win_x, win_y = window.sliding_window_idx(imgs[0])
    # apply the coordinates on all images
    window_stack = window.multi_sliding_window_array(imgs, win_x, win_y, swap_time_dim=False)
    return np.float32(window_stack)


@pytest.fixture()
def imgs_win_stack(imgs_win):
    """Prepare a stack of images from first img win with some random noise."""
    imgs_win_ = imgs_win
    for _ in range(10):
        imgs_wins = np.concatenate(
            [
                imgs_win,
                # np.float32(imgs_win_ * np.random.rand(*imgs_win_.shape)),
                imgs_win_ * np.random.rand(*imgs_win_.shape).astype(np.float32),
            ],
            axis=0,
        )
    return imgs_wins


@pytest.fixture()
def mask(imgs_win):
    """Prepare a mask array fpr surrounding areas of interrogation window."""
    mask_array = np.ones((imgs_win.shape[-2], imgs_win.shape[-1]), dtype=np.float32)
    mask_array = np.repeat(np.expand_dims(mask_array, 0), imgs_win.shape[1], axis=0)
    return mask_array



"""Numba just in time compiling cross-correlation related functions."""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def rfft2(x):
    """JIT variant of rfft2."""
    return np.fft.rfft2(x)


@nb.njit(cache=True)
def rfft2_x(x, fsize):
    """JIT variant of rfft2 with fsize parameter."""
    return np.fft.rfft2(x, fsize)


@nb.njit(cache=True)
def irfft2(x):
    """JIT variant of irfft2."""
    return np.fft.irfft2(x)


@nb.njit(cache=True)
def fftshift(x, axes):
    """JIT variant of fftshift."""
    return np.fft.fftshift(x, axes)


@nb.njit(cache=True)
def conj(x):
    """JIT variant of conj."""
    return np.conj(x)


@nb.njit(nb.float32[:, :](nb.float32[:, :]), cache=True)
def normalize_intensity(img: np.ndarray) -> np.ndarray:
    """Normalize intensity of an image interrogation window using numba back-end.

    Parameters
    ----------
    img : np.ndarray (y, x)
        Image window

    Returns
    -------
    np.ndarray
        (y, x) array with normalized intensities of window

    """
    img_mean = np.float32(np.mean(img))
    img = img - img_mean
    img_std = np.float32(np.std(img))
    if img_std != 0:
        img = img / img_std
    else:
        img = np.zeros_like(img, dtype=np.float32)
    return np.clip(img, 0, img.max())


@nb.njit(
    nb.float32[:, :, :](nb.float32[:, :, :], nb.float32[:, :, :]),
    parallel=True,
    nogil=True,
    cache=True,
)
def ncc(image_a, image_b):
    """Perform normalized cross correlation performed on a set of interrogation window pairs with numba back-end.

    Parameters
    ----------
    image_a : np.ndarray
        uint8 type array (w, y, x) containing a single image, sliced into interrogation windows (w)
    image_b : np.ndarray
        uint8 type array (w, y, x) containing the next image, sliced into interrogation windows (w)

    Returns
    -------
    np.ndarray
        float64 (w, y, x) correlations of interrogation window pixels

    """
    res = np.empty_like(image_a, dtype=nb.float32)
    const = np.multiply(*image_a.shape[-2:])
    for n in nb.prange(image_a.shape[0]):
        ima = image_a[n]
        imb = image_b[n]
        ima = normalize_intensity(ima)
        imb = normalize_intensity(imb)
        f2a = conj(rfft2(ima))
        f2b = rfft2(imb)
        corr = fftshift(irfft2(f2a * f2b).real, axes=(-2, -1))
        res[n] = np.clip(corr / const, 0, 1)
    return res


@nb.njit()
def slice_a_b(imgs, n, mask, idx):
    """Extract one frame as source and the next as destination for image velocimetry.

    This function masks non-relevant areas in the source image and removes windows that are not relevant.

    Parameters
    ----------
    imgs : np.ndarray
        (i, w, y, x) set of images (i), subdivided into windows (w) for cross-correlation computation.
    n : int
        index of imgs to extract. n + 1 will be extracted as next frame
    mask : np.ndarray
        (y, x) array containing ones in the area covered by a window, and zeros in the search area around the window.
    idx : np.ndarray, optional
        contains which windows (dimension w in imgs) should be cross correlated. If not provided, all windows are
        treated.

    """
    return imgs[n, idx] * mask[idx], imgs[n + 1, idx]


@nb.njit(
    nb.float32[:, :, :, :](nb.float32[:, :, :, :], nb.float32[:, :, :], nb.boolean[:]),
    cache=True,
    parallel=True,
    nogil=True,
)
def multi_img_ncc(imgs, mask, idx):
    """Compute correlation over all image pairs in `imgs` using numba back-end.

    Correlations are computed for each interrogation window (dim1) and each image pair (dim0)
    Because pair-wise correlation is performed the resulting dim0 size one stride smaller than the input imgs array.

    Parameters
    ----------
    imgs : np.ndarray
        (i, w, y, x) set of images (i), subdivided into windows (w) for cross-correlation computation.
    mask : np.ndarray
        (y, x) array containing ones in the area covered by a window, and zeros in the search area around the window.
    idx : np.ndarray, optional
        contains which windows (dimension w in imgs) should be cross correlated. If not provided, all windows are
        treated.

    Returns
    -------
    np.ndarray
        float64 (i - 1, w, y, x) correlations of interrogation window pixels for each image pair spanning i.

    """
    corr = np.empty(
        (len(imgs) - 1, imgs.shape[-3], imgs.shape[-2], imgs.shape[-1]),
        dtype=nb.float32,
    )
    corr.fill(np.nan)
    for n in nb.prange(len(imgs) - 1):
        img_a, img_b = slice_a_b(imgs, n, mask, idx)
        corr[n, idx] = ncc(img_a, img_b).astype(nb.float32)
    return corr


@nb.njit(cache=True)
def peak_position(corr):
    """Compute peak positions for correlations in each interrogation window using numba back-end."""
    eps = 1e-7
    idx = np.argmax(corr)
    peak1_i, peak1_j = idx // len(corr), idx % len(corr)
    # check if valid
    valid = peak1_i != 0 and peak1_i != corr.shape[-2] - 1 and peak1_j != 0 and peak1_j != corr.shape[-1] - 1
    if valid:
        corr = corr + eps  # prevents log(0) = nan if "gaussian" is used (notebook)
        c = corr[peak1_i, peak1_j] + eps
        cl = corr[peak1_i - 1, peak1_j] + eps
        cr = corr[peak1_i + 1, peak1_j] + eps
        cd = corr[peak1_i, peak1_j - 1] + eps
        cu = corr[peak1_i, peak1_j + 1] + eps

        # gaussian peak
        nom1 = np.log(cl) - np.log(cr)
        den1 = 2 * np.log(cl) - 4 * np.log(c) + 2 * np.log(cr) + eps
        nom2 = np.log(cd) - np.log(cu)
        den2 = 2 * np.log(cd) - 4 * np.log(c) + 2 * np.log(cu) + eps

        subp_peak_position = np.array([peak1_i + nom1 / den1, peak1_j + nom2 / den2])
    else:
        subp_peak_position = np.array([np.nan, np.nan])
    return subp_peak_position


@nb.njit(parallel=True, cache=True)
def u_v_displacement(
    corr,
    n_rows,
    n_cols,
):
    """Compute u (x-direction) and v (y-direction) displacements.

    u and v displacements are computed from correlations in windows and number and rows / columns using numba
    back-end.

    Parameters
    ----------
    corr : np.ndarray
        (w, y, x) correlation planes for each interrogation window (w).
    n_rows : int
        number of rows in the correlation map.
    n_cols : int
        number of columns in the correlation map.

    Returns
    -------
    u : np.ndarray
        (n_rows, n_cols) array of x-direction velocimetry results in pixel displacements.
    v : np.ndarray
        (n_rows, n_cols) array of y-direction velocimetry results in pixel displacements.

    """
    u = np.zeros((n_rows, n_cols))
    v = np.zeros((n_rows, n_cols))

    # center point of the correlation map
    default_peak_position = np.floor(np.array(corr[0, :, :].shape) / 2)
    for k in nb.prange(n_rows):
        for m in nb.prange(n_cols):
            peak = (
                peak_position(
                    corr[k * n_cols + m],
                )
                - default_peak_position
            )
            u[k, m] = peak[1]
            v[k, m] = peak[0]
    return u, v


@nb.njit(parallel=True, cache=True)
def multi_u_v_displacement(
    corr,
    n_rows,
    n_cols,
):
    """Compute u and v displacement for multiple images at once.

    Parameters
    ----------
    corr : np.ndarray
        (i, w, y, x) correlations for each interrogation window (w) in each image pair (i).
    n_rows : int
        number of rows in end result
    n_cols : int
        number of columns in end result

    Returns
    -------
    u : np.ndarray
        Stack of x-direction velocimetry results (i, Y, X) in pixel displacements.
    v : np.ndarray
        Stack of y-direction velocimetry results (i, Y, X) in pixel displacements.

    """
    u = np.zeros((len(corr), n_rows, n_cols))
    v = np.zeros((len(corr), n_rows, n_cols))
    for i in nb.prange(len(corr)):
        _u, _v = u_v_displacement(corr[i], n_rows, n_cols)
        u[i] = _u
        v[i] = _v
    return u, v


@nb.njit(nb.float64[:](nb.float64[:, :, :]), parallel=True, cache=True, nogil=True)
def signal_to_noise(corr: np.ndarray):
    """Compute signal-to-noise ratio per interrogation window.

    This is computed by dividing the peak of the correlation field by the mean.

    Parameters
    ----------
    corr : np.ndarray
        (w, y, x) correlations for each interrogation window (w).

    Returns
    -------
    np.ndarray
        vector with signal to noise ratios for each window.

    """
    s2n = np.empty(
        len(corr),
        dtype=nb.float64,
    )
    for n in nb.prange(len(corr)):
        _c = corr[n]
        _c_max = _c.max()
        if _c_max < 1e-3:
            # no correlation
            s2n[n] = 0.0
        else:
            s2n[n] = _c_max / abs(_c.mean())
    return s2n


@nb.njit(nb.float64[:, :](nb.float64[:, :, :, :]), parallel=True, cache=True, nogil=True)
def multi_signal_to_noise(corr: np.ndarray):
    """Compute signal to noise for multiple images at once."""
    s2n = np.zeros(corr.shape[0:2])
    for i in nb.prange(len(corr)):
        _s2n = signal_to_noise(corr[i])
        s2n[i] = _s2n
    return s2n



import time

import numpy as np
import pytest

import ffpiv.pnb as pnb
import ffpiv.pnp as pnp
from ffpiv import window


@pytest.fixture()
def img_pair(imgs_win):
    # only return image 0 and 1
    img_pair = imgs_win[0:2]
    return img_pair


@pytest.fixture()
def dims(imgs):
    x, y = window.get_rect_coordinates(
        dim_size=imgs.shape[-2:],
        window_size=(64, 64),
        overlap=(32, 32),
    )
    nrows, ncols = len(y), len(x)
    return nrows, ncols


@pytest.fixture()
def correlations(img_pair):
    corrs = pnb.ncc(*img_pair)
    return corrs * np.random.rand(*corrs.shape) * 0.005


def test_ncc(img_pair):
    """Test correlation analysis on a pair of image windows."""
    image_a, image_b = img_pair
    t1 = time.time()
    res_nb = pnb.ncc(image_a, image_b)
    t2 = time.time()
    time_nb = t2 - t1
    print(f"Numba took {time_nb} secs.")
    t1 = time.time()
    res_np = pnp.ncc(image_a, image_b)
    t2 = time.time()
    time_np = t2 - t1
    print(f"Numpy took {time_np} secs.")
    assert np.allclose(res_nb, res_np, atol=1e-6, rtol=1e-5)
    # plt.imshow(res_nb[0])
    # plt.colorbar()
    # plt.show()
    # TODO: also test if values are close to expected values


def test_multi_img_ncc(imgs_win_stack, mask):
    """Test cross correlation with several hundreds of images."""
    t1 = time.time()
    idx = np.repeat(True, imgs_win_stack.shape[-3])
    res_nb = pnb.multi_img_ncc(imgs_win_stack, mask, idx)
    t2 = time.time()
    time_nb = t2 - t1
    print(f"Numba took {time_nb} secs.")
    t1 = time.time()
    res_np = pnp.multi_img_ncc(imgs_win_stack, mask)
    t2 = time.time()
    time_nb = t2 - t1
    print(f"Numpy took {time_nb} secs.")
    assert np.allclose(res_nb, res_np, atol=1e-6, rtol=1e-5)


def test_u_v_displacement(correlations, dims):
    """Test displacement functionalities."""
    n_rows, n_cols = dims
    t1 = time.time()
    _ = pnb.u_v_displacement(correlations, n_rows, n_cols)
    t2 = time.time()
    print(f"Peak position search took {t2 - t1} seconds")

    t1 = time.time()
    _ = pnp.u_v_displacement(correlations, n_rows, n_cols)
    t2 = time.time()
    print(f"Peak position search with OpenPIV took {t2 - t1} seconds")

    # plt.quiver(u2, v2, color="r", alpha=0.5)
    # plt.quiver(u, v, color="b", alpha=0.5)
    # plt.show()


def test_peaks_numpy(correlations):
    peaks = pnp.peak_position(correlations)
    print(peaks)


def test_signal_to_noise(correlations):
    # compile
    _ = pnb.signal_to_noise(correlations)
    t1 = time.time()
    _ = pnb.signal_to_noise(correlations)
    t2 = time.time()
    print(f"Signal to noise calculation took {t2 - t1} seconds")



import numpy as np

from ffpiv import piv, piv_stack


def test_piv_stack(imgs):
    """Test image stack result in one go."""
    u, v = piv_stack(imgs, (64, 64), (32, 32))

    # Assertions to validate the outputs
    assert isinstance(u, np.ndarray), "Expected u to be a numpy array"
    assert isinstance(v, np.ndarray), "Expected v to be a numpy array"

    # Additional assertions based on expected shapes
    assert u.shape == v.shape, "Expected u and v to have the same shape"
    assert len(u.shape) == 3, "Expected u and v to have 3 dimensions"
    assert u.shape[0] == len(imgs) - 1, "Expected the first dimension of u to match the number of images minus 1"
    assert v.shape[0] == len(imgs) - 1, "Expected the first dimension of v to match the number of images minus 1"


def test_piv(imgs):
    """Test single piv result for image pair."""
    img_a = imgs[0]
    img_b = imgs[1]
    u, v = piv(img_a, img_b, window_size=(64, 64), overlap=(32, 16))
    assert isinstance(u, np.ndarray)
    assert isinstance(v, np.ndarray)
    assert u.shape == v.shape



"""Interfacing wrapper functions to disclose functionalities."""

import gc
import time
import warnings
from typing import Literal, Optional, Tuple

import numpy as np

import ffpiv.pnb as pnb
import ffpiv.pnp as pnp
from ffpiv import window


def subwindows(
    imgs: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    search_area_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (0, 0),
):
    """Subdivide image stack into windows with associated coordinates of center."""
    x, y = window.get_rect_coordinates(
        dim_size=imgs.shape[-2:], window_size=window_size, overlap=overlap, search_area_size=search_area_size
    )

    win_x, win_y = window.sliding_window_idx(
        imgs[0],
        window_size=window_size,
        search_area_size=search_area_size,
        overlap=overlap,
    )
    window_stack = window.multi_sliding_window_array(imgs, win_x, win_y)
    return x, y, window_stack


def coords(
    dim_size: Tuple[int, int],
    window_size: Tuple[int, int],
    overlap: Tuple[int, int],
    search_area_size: Optional[Tuple[int, int]] = None,
    center_on_field: bool = False,
):
    """Create coordinates (x, y) of velocimetry results.

    Overlap can be provided in case each interrogation window is to overlap with the neighbouring interrogation window.

    Parameters
    ----------
    dim_size : Tuple[int, int]
        size of the ingoing images (y, x)
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension.
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x( second) dimension.
    search_area_size : tuple[int, int], optional
        Search area window size in y (first) and x (second) dimension. If not provided, set to window_size
    center_on_field : bool, optional
        whether the center of interrogation window is returned (True) or (False) the bottom left (default=True)

    Returns
    -------
    x, y: np.ndarray (1D), np.ndarray (1D)
        x- and y-coordinates in axis form

    """
    return window.get_rect_coordinates(
        dim_size=dim_size,
        window_size=window_size,
        overlap=overlap,
        search_area_size=search_area_size,
        center_on_field=center_on_field,
    )


def piv(
    img_a: np.ndarray,
    img_b: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (0, 0),
    engine: Literal["numba", "numpy"] = "numba",
):
    """Perform particle image velocimetry on a pair of images.

    Parameters
    ----------
    img_a : np.ndarray
        First image.
    img_b : np.ndarray
        Second image.
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension.
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x( second) dimension.
    engine : Literal["numba", "numpy"], optional
        Compute correlations and displacements with "numba" (default) or "numpy"

    Returns
    -------
    u : np.ndarray
        X-direction velocimetry results in pixel displacements.
    v : np.ndarray
        Y-direction velocimetry results in pixel displacements.

    """
    # get subwindows
    imgs = np.stack((img_a, img_b), axis=0).astype(np.float64)
    # get correlations and row/column layout
    x, y, corr = cross_corr(imgs, window_size=window_size, overlap=overlap, engine=engine)
    # get displacements
    n_rows, n_cols = len(y), len(x)
    u, v = u_v_displacement(corr, n_rows, n_cols)
    return u[0], v[0]


def piv_stack(
    imgs: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (0, 0),
    engine: Literal["numba", "numpy"] = "numba",
):
    """Perform particle image velocimetry over a stack of images.

    Parameters
    ----------
    imgs : np.ndarray
        Stack of images [i * y * x]
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x( second) dimension
    engine : Literal["numba", "numpy"], optional
        Compute correlations and displacements with "numba" (default) or "numpy"

    Returns
    -------
    u : np.ndarray
        Stack of x-direction velocimetry results (i -1, Y, X) in pixel displacements.
    v : np.ndarray
        Stack of y-direction velocimetry results (i -1, Y, X) in pixel displacements.

    """
    # get correlations and row/column layout
    x, y, corr = cross_corr(imgs, window_size=window_size, overlap=overlap, engine=engine)
    # get displacements
    n_rows, n_cols = len(y), len(x)
    if engine == "numpy":
        u, v = pnp.multi_u_v_displacement(corr, n_rows, n_cols)
    else:
        u, v = pnb.multi_u_v_displacement(corr, n_rows, n_cols)
    return u, v


def cross_corr(
    imgs: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (32, 32),
    search_area_size: Optional[Tuple[int, int]] = None,
    engine: Literal["numba", "numpy"] = "numba",
    normalize: bool = False,
    verbose: bool = True,
):
    """Compute correlations over a stack of images using interrogation windows.

    Parameters
    ----------
    imgs : np.ndarray
        Stack of images [i * y * x]
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension.
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x (second) dimension.
    search_area_size : tuple[int, int], optional
        size of the search area. This is used in the second frame window set, to explore areas larger than
        `window_size`.
    engine : Literal["numba", "numpy"], optional
        The engine to use for calculation, by default "numba".
    normalize : bool, optional
        if set, each window will be normalized with spatial mean and standard deviation, and numbers capped to 0.
    verbose : bool, optional
        if set (default), warnings will be displayed if the amount of available memory is low.

    Returns
    -------
    n_rows : int
        The number of rows of windows.
    n_cols : int
        The number of columns of windows.
    corr : np.ndarray
        A 4D array containing per image and per interrogation window the correlation results.

    """
    # check if masking is needed
    window_size = window.round_to_even(window_size)
    if search_area_size is None:
        search_area_size = window_size
    # check if search_area_size contains uneven numbers
    search_area_size = window.round_to_even(search_area_size)
    # search_area_size must be at least equal to the window size
    search_area_size = max(search_area_size, window_size)

    # estimate amount of memory required for intermediate in-memory storage
    dim_size = (imgs.shape[-2], imgs.shape[-1])
    # win_shape = window.get_array_shape(
    #     dim_size=(imgs.shape[-2], imgs.shape[-1]),
    #     window_size=window_size,
    #     overlap=overlap,
    #     search_area_size=search_area_size
    # )
    # var_size1 = len(imgs) * win_shape[0] * win_shape[1] * search_area_size[0] * search_area_size[1] * 4 / 1e9
    # Prepare subwindows
    imgs = np.float32(imgs)
    req_mem = window.required_memory(
        len(imgs), dim_size=dim_size, window_size=window_size, overlap=overlap, search_area_size=search_area_size
    )
    avail_mem = window.available_memory()
    if verbose:
        if avail_mem - req_mem < 0:
            warnings.warn(
                f"You may have too little physical memory ({avail_mem / 1e9} GB) available for this problem. "
                f"You may need {req_mem / 1e9} GB. ffpiv may slow down or crash! Reduce the amount of frames "
                f" interpreted in one go.",
                stacklevel=2,
            )
            # wait for a while so that user can read the message
            time.sleep(1)
    x, y, window_stack = subwindows(
        imgs,
        window_size=window_size,
        search_area_size=search_area_size,
        overlap=overlap,
    )
    # normalization
    if normalize:
        window_stack = window.normalize(window_stack, mode="xy")
    # prepare a mask for the first frame of analysis
    mask = window.mask_search_area(window_size=window_size, search_area_size=search_area_size)
    # expand mask over total amount of sub windows
    mask = np.repeat(np.expand_dims(mask, 0), window_stack.shape[1], axis=0).astype(np.float32)

    # fully missing should be ignored
    idx = np.any(window_stack[0] != 0, axis=(-1, -2))
    if engine == "numpy":
        corr = pnp.multi_img_ncc(window_stack, mask=mask, idx=idx)
    else:
        corr = pnb.multi_img_ncc(window_stack, mask=mask, idx=idx)
    # memory cleanup
    del idx, mask, window_stack
    gc.collect()
    return x, y, corr


def u_v_displacement(corr: np.array, n_rows: int, n_cols: int, engine: Literal["numba", "numpy"] = "numba"):
    """Compute x-direction and y-directional displacements.

    Parameters
    ----------
    corr : np.array
        4D array [i, w, x, y] cross correlations computed per image and interrogation window pixel.
    n_rows : int
        The number of rows in the output displacement arrays.
    n_cols : int
        The number of columns in the output displacement arrays.
    engine : Literal["numba", "numpy"], optional
        The computational engine to use for calculating displacements, either "numba" or "numpy". Default is "numba".

    Returns
    -------
    u : np.array
        An array containing the u-component of the displacements.
    v : np.array
        An array containing the v-component of the displacements.

    """
    if engine == "numpy":
        u, v = pnp.multi_u_v_displacement(corr, n_rows, n_cols)
    else:
        u, v = pnb.multi_u_v_displacement(corr, n_rows, n_cols)
    return u, v



"""Numpy cross-correlation related functions."""

import numpy as np


def normalize_intensity(img: np.uint8) -> np.float64:
    """Normalize intensity of an image interrogation window using numpy back-end.

    Parameters
    ----------
    img : np.ndarray (w * y * x)
        Image subdivided into interrogation windows (w)

    Returns
    -------
    np.ndarray
        [w * y * z] array with normalized intensities per window

    """
    img = img.astype(np.float32)
    img_mean = img.mean(axis=(-2, -1), keepdims=True)
    img = img - img_mean
    img_std = img.std(axis=(-2, -1), keepdims=True)
    img = np.divide(img, img_std, out=np.zeros_like(img), where=(img_std != 0))
    return np.clip(img, 0, img.max())


def ncc(image_a, image_b):
    """Perform normalized cross correlation performed on a set of interrogation window pairs with numpy back-end.

    Parameters
    ----------
    image_a : np.ndarray
        uint8 type array [w, y, x] containing a single image, sliced into interrogation windows (w)
    image_b : np.ndarray
        uint8 type array [w, y, x] containing the next image, sliced into interrogation windows (w)

    Returns
    -------
    np.ndarray
        float64 [w * y * x] correlations of interrogation window pixels

    """
    const = np.multiply(*image_a.shape[-2:])
    image_a = normalize_intensity(image_a)
    image_b = normalize_intensity(image_b)
    f2a = np.conj(np.fft.rfft2(image_a))
    f2b = np.fft.rfft2(image_b)
    return np.clip(np.fft.fftshift(np.fft.irfft2(f2a * f2b).real, axes=(-2, -1)) / const, 0, 1)


def multi_img_ncc(imgs, mask=None, idx=None):
    """Compute correlation over all image pairs in `imgs` using numpy back-end.

    Correlations are computed for each interrogation window (dim1) and each image pair (dim0)
    Because pair-wise correlation is performed the resulting dim0 size one stride smaller than the input imgs array.

    Parameters
    ----------
    imgs : np.ndarray
        [i, w, y, x] set of images (i), subdivided into windows (w) for cross-correlation computation.
    mask : np.ndarray
        [y, x] array containing ones in the area covered by a window, and zeros in the search area around the window.
    idx : np.ndarray, optional
        contains which windows (dimension w in imgs) should be cross correlated. If not provided, all windows are
        treated.

    Returns
    -------
    np.ndarray
        float64 [(i - 1), w, y, x] correlations of interrogation window pixels for each image pair spanning i.

    """
    corr = np.empty((len(imgs) - 1, imgs.shape[-3], imgs.shape[-2], imgs.shape[-1]), dtype=np.float32)
    corr.fill(np.nan)
    if idx is None:
        idx = np.repeat(True, imgs.shape[-3])
    # mask = np.expand_dims(mask, 0)
    for n in range(len(imgs) - 1):
        img_a = imgs[n, idx] * mask[idx]
        img_b = imgs[n + 1, idx]
        # import matplotlib.pyplot as plt
        # f, axs = plt.subplots(ncols=2)
        # p1 = axs[0].imshow(img_a[50], cmap="Greys_r");plt.colorbar(p1)
        # p2 = axs[1].imshow(img_b[50], cmap="Greys_r");plt.colorbar(p2)
        # plt.show()
        res = ncc(img_a, img_b)
        # import matplotlib.pyplot as plt
        # p1 = plt.imshow(res[68]);plt.colorbar(p1)
        # plt.show()
        corr[n, idx] = res.astype(np.float32)
    return corr


def peak_position(corr):
    """Compute peak positions for correlations in each interrogation window using numpy back-end.

    Parameters
    ----------
    corr : np.ndarray
        A 3D array of shape (w, y, x) representing the correlation maps for which the peak positions are to be
        identified.

    Returns
    -------
    subp_peak_position : np.ndarray
        A 2D array of shape (w, 2) where each row corresponds to the subpixel peak positions (i, j) for each 2D slice
        in the input correlation maps.

    """
    eps = 1e-7
    # pre-define sub peak position array
    subp_peak_position = np.zeros((len(corr), 2)) * np.nan
    # Find argmax along axis (1, 2) for each 2D slice of the input
    idx = np.argmax(corr.reshape(corr.shape[0], -1), axis=1)
    peak1_i = idx // corr.shape[1]
    peak1_j = idx % corr.shape[2]

    # Adding eps to avoid log(0)
    corr = corr + eps
    valid_idx = np.where(
        np.all(
            np.array([peak1_i != 0, peak1_i != corr.shape[-2] - 1, peak1_j != 0, peak1_j != corr.shape[-1] - 1]), axis=0
        )
    )
    peak1_i = peak1_i[valid_idx]
    peak1_j = peak1_j[valid_idx]

    # Indexing the peak and neighboring points for vectorized operations
    c = corr[valid_idx, peak1_i, peak1_j] + eps
    cl = corr[valid_idx, peak1_i - 1, peak1_j] + eps
    cr = corr[valid_idx, peak1_i + 1, peak1_j] + eps
    cd = corr[valid_idx, peak1_i, peak1_j - 1] + eps
    cu = corr[valid_idx, peak1_i, peak1_j + 1] + eps

    # Gaussian peak calculations (nom1, den1, nom2, den2)
    nom1 = np.log(cl) - np.log(cr)
    den1 = 2 * np.log(cl) - 4 * np.log(c) + 2 * np.log(cr) + eps
    nom2 = np.log(cd) - np.log(cu)
    den2 = 2 * np.log(cd) - 4 * np.log(c) + 2 * np.log(cu) + eps

    # Subpixel peak position
    subp_peak_position[valid_idx] = np.vstack([peak1_i + nom1 / den1, peak1_j + nom2 / den2]).T

    return subp_peak_position


def u_v_displacement(corr: np.ndarray, n_rows: int, n_cols: int):
    """Compute u (x-direction) and v (y-direction) displacements from correlations using numpy back end.

    Results are organized following the row and column organization provided by `n_rows` and `n_cols`.

    Parameters
    ----------
    corr : np.ndarray
        A 3D array of shape [w, y, x] containing the correlation maps.
    n_rows : int
        Number of rows in the image interrogation windows.
    n_cols : int
        Number of columns in the image interrogation windows.

    Returns
    -------
    u : np.ndarray
        2D array containing x-direction displacements for each image pair and window.
    v : np.ndarray
        2D array containing y-direction displacements for each image pair and window.

    """
    peaks = peak_position(corr.astype(np.float64))
    peaks_def = np.floor(np.array(corr[0, :, :].shape) / 2)
    u = peaks[:, 1].reshape(n_rows, n_cols) - peaks_def[1]
    v = peaks[:, 0].reshape(n_rows, n_cols) - peaks_def[0]
    return u, v


def multi_u_v_displacement(corr: np.ndarray, n_rows: int, n_cols: int):
    """Compute u (x-direction) and v (y-direction) displacements for multiple correlation windows using numpy back-end.

    Parameters
    ----------
    corr : np.ndarray
        A 4D array of shape [(i - 1), w, y, x] containing the correlation maps.
    n_rows : int
        Number of rows in the image interrogation windows.
    n_cols : int
        Number of columns in the image interrogation windows.

    Returns
    -------
    u : np.ndarray
        3D array containing x-direction displacements for each image pair and window.
    v : np.ndarray
        3D array containing y-direction displacements for each image pair and window.

    """
    u = np.empty((corr.shape[0], n_rows, n_cols))
    v = np.empty((corr.shape[0], n_rows, n_cols))
    for i in range(corr.shape[0]):
        u[i], v[i] = u_v_displacement(corr[i], n_rows, n_cols)
    return u, v
